﻿Imports System
Namespace LogisticMultiClassGradient
	Friend Class LogisticMultiProgram
		Shared Sub Main(ByVal args() As String)
			Console.WriteLine(vbLf & "Begin multi-class logistic regression classification demo")

			Dim numFeatures As Integer = 4
			Dim numClasses As Integer = 3
			Dim numRows As Integer = 1000
			Console.WriteLine(vbLf & "Generating " & numRows & " artificial data items with " & numFeatures & " features")
			Dim data()() As Double = MakeDummyData(numFeatures, numClasses, numRows, 0)

			Console.WriteLine("Splitting data to train (80%) and test matrices")
			Dim trainData()() As Double
			Dim testData()() As Double
			SplitTrainTest(data, 0.80, 7, trainData, testData)
			Console.WriteLine("Done")

			Console.WriteLine(vbLf & "Training data: " & vbLf)
			ShowData(trainData, 3, 2, True)

			Console.WriteLine(vbLf & "Test data: " & vbLf)
			ShowData(testData, 3, 2, True)

			Console.WriteLine("Creating multi-class LR classifier")
			Dim lc As New LogisticMulti(numFeatures, numClasses)

			Dim maxEpochs As Integer = 100
			Console.WriteLine("Setting training maxEpochs = " & maxEpochs)
			Dim learnRate As Double = 0.01
			Console.WriteLine("Setting learning rate      = " & learnRate.ToString("F2"))
			Dim decay As Double = 0.10
			Console.WriteLine("Setting weight decay       = " & decay.ToString("F2"))

			Console.WriteLine(vbLf & "Starting training using (batch) gradient descent" & vbLf)
			lc.Train(trainData, maxEpochs, learnRate, decay)
			Console.WriteLine(vbLf & "Done" & vbLf)

			Dim bestWts()() As Double = lc.GetWeights()
			Dim bestBiases() As Double = lc.GetBiases()

			Console.WriteLine("Best weights found:")
			ShowData(bestWts, bestWts.Length, 3, True)

			Console.WriteLine("Best biases found:")
			ShowVector(bestBiases, 3, True)

			Dim trainAcc As Double = lc.Accuracy(trainData)
			Console.WriteLine("Prediction accuracy on training data = " & trainAcc.ToString("F4"))

			Dim testAcc As Double = lc.Accuracy(testData)
			Console.WriteLine("Prediction accuracy on test data     = " & testAcc.ToString("F4"))

			Console.WriteLine(vbLf & "End demo" & vbLf)
			Console.ReadLine()
		End Sub ' Main

		Public Shared Sub ShowData(ByVal data()() As Double, ByVal numRows As Integer, ByVal decimals As Integer, ByVal indices As Boolean)
			Dim len As Integer = data.Length.ToString().Length
			For i As Integer = 0 To numRows - 1
				If indices = True Then
					Console.Write("[" & i.ToString().PadLeft(len) & "]  ")
				End If
				For j As Integer = 0 To data(i).Length - 1
					Dim v As Double = data(i)(j)
					If v >= 0.0 Then
						Console.Write(" ") ' '+'
					End If
					Console.Write(v.ToString("F" & decimals) & "  ")
				Next j
				Console.WriteLine("")
			Next i

			If numRows < data.Length Then
				Console.WriteLine(". . .")
				Dim lastRow As Integer = data.Length - 1
				If indices = True Then
					Console.Write("[" & lastRow.ToString().PadLeft(len) & "]  ")
				End If
				For j As Integer = 0 To data(lastRow).Length - 1
					Dim v As Double = data(lastRow)(j)
					If v >= 0.0 Then
						Console.Write(" ") ' '+'
					End If
					Console.Write(v.ToString("F" & decimals) & "  ")
				Next j
			End If
			Console.WriteLine(vbLf)
		End Sub

		Public Shared Sub ShowVector(ByVal vector() As Double, ByVal decimals As Integer, ByVal newLine As Boolean)
			For i As Integer = 0 To vector.Length - 1
				Console.Write(vector(i).ToString("F" & decimals) & " ")
			Next i
			Console.WriteLine("")
			If newLine = True Then
				Console.WriteLine("")
			End If
		End Sub

		Private Shared Function MakeDummyData(ByVal numFeatures As Integer, ByVal numClasses As Integer, ByVal numRows As Integer, ByVal seed As Integer) As Double()()
			Dim rnd As New Random(seed) ' make random wts and biases
			Dim wts(numFeatures - 1)() As Double
			For i As Integer = 0 To numFeatures - 1
				wts(i) = New Double(numClasses - 1){}
			Next i
			Dim hi As Double = 10.0
			Dim lo As Double = -10.0
			For i As Integer = 0 To numFeatures - 1
				For j As Integer = 0 To numClasses - 1
					wts(i)(j) = (hi - lo) * rnd.NextDouble() + lo
				Next j
			Next i
			Dim biases(numClasses - 1) As Double
			For i As Integer = 0 To numClasses - 1
				biases(i) = (hi - lo) * rnd.NextDouble() + lo
			Next i

			Console.WriteLine("Generating weights are: ")
			ShowData(wts, wts.Length, 2, True)
			Console.WriteLine("Generating biases are: ")
			ShowVector(biases, 2, True)

			Dim result(numRows - 1)() As Double ' allocate result
			For i As Integer = 0 To numRows - 1
				result(i) = New Double(numFeatures + numClasses - 1){}
			Next i

			For i As Integer = 0 To numRows - 1 ' create one row at a time
				Dim x(numFeatures - 1) As Double ' generate random x-values
				For j As Integer = 0 To numFeatures - 1
					x(j) = (hi - lo) * rnd.NextDouble() + lo
				Next j

				Dim y(numClasses - 1) As Double ' computed outputs storage
				For j As Integer = 0 To numClasses - 1 ' compute z-values
					For f As Integer = 0 To numFeatures - 1
						y(j) += x(f) * wts(f)(j)
					Next f
					y(j) += biases(j)
				Next j

				' determine loc. of max (no need for 1 / 1 + e^-z)
				Dim maxIndex As Integer = 0
				Dim maxVal As Double = y(0)
				For c As Integer = 0 To numClasses - 1
					If y(c) > maxVal Then
						maxVal = y(c)
						maxIndex = c
					End If
				Next c

				For c As Integer = 0 To numClasses - 1 ' convert y to 0s or 1s
					If c = maxIndex Then
						y(c) = 1.0
					Else
						y(c) = 0.0
					End If
				Next c

				Dim col As Integer = 0 ' copy x and y into result
				For f As Integer = 0 To numFeatures - 1
					result(i)(col) = x(f)
					col += 1
				Next f
				For c As Integer = 0 To numClasses - 1
					result(i)(col) = y(c)
					col += 1
				Next c
			Next i
			Return result
		End Function

		Private Shared Sub SplitTrainTest(ByVal allData()() As Double, ByVal trainPct As Double, ByVal seed As Integer, <System.Runtime.InteropServices.Out()> ByRef trainData()() As Double, <System.Runtime.InteropServices.Out()> ByRef testData()() As Double)
			Dim rnd As New Random(seed)
			Dim totRows As Integer = allData.Length
			Dim numTrainRows As Integer = CInt(Fix(totRows * trainPct)) ' typically 80%
			Dim numTestRows As Integer = totRows - numTrainRows
			trainData = New Double(numTrainRows - 1)(){}
			testData = New Double(numTestRows - 1)(){}

			Dim copy(allData.Length - 1)() As Double ' ref copy of all data
			For i As Integer = 0 To copy.Length - 1
				copy(i) = allData(i)
			Next i

			For i As Integer = 0 To copy.Length - 1 ' scramble order
				Dim r As Integer = rnd.Next(i, copy.Length) ' use Fisher-Yates
				Dim tmp() As Double = copy(r)
				copy(r) = copy(i)
				copy(i) = tmp
			Next i
			For i As Integer = 0 To numTrainRows - 1
				trainData(i) = copy(i)
			Next i

			For i As Integer = 0 To numTestRows - 1
				testData(i) = copy(i + numTrainRows)
			Next i
		End Sub ' MakeTrainTest


	End Class ' Program

	Public Class LogisticMulti
		Private numFeatures As Integer
		Private numClasses As Integer
		Private weights()() As Double ' [feature][class]
		Private biases() As Double ' [class]

		Public Sub New(ByVal numFeatures As Integer, ByVal numClasses As Integer)
			Me.numFeatures = numFeatures
			Me.numClasses = numClasses
			Me.weights = MakeMatrix(numFeatures, numClasses)
			Me.biases = New Double(numClasses - 1){}
		End Sub

		Private Function MakeMatrix(ByVal rows As Integer, ByVal cols As Integer) As Double()()
			Dim result(rows - 1)() As Double
			For i As Integer = 0 To rows - 1
				result(i) = New Double(cols - 1){}
			Next i
			Return result
		End Function

		Public Sub SetWeights(ByVal wts()() As Double, ByVal b() As Double)
			' set weights[][] and biases[]
			For i As Integer = 0 To numFeatures - 1
				For j As Integer = 0 To numClasses - 1
					Me.weights(i)(j) = wts(i)(j)
				Next j
			Next i
			For j As Integer = 0 To numClasses - 1
				Me.biases(j) = b(j)
			Next j
		End Sub

		Public Function GetWeights() As Double()()
			Dim result(numFeatures - 1)() As Double
			For i As Integer = 0 To numFeatures - 1
				result(i) = New Double(numClasses - 1){}
			Next i
			For i As Integer = 0 To numFeatures - 1
				For j As Integer = 0 To numClasses - 1
					result(i)(j) = Me.weights(i)(j)
				Next j
			Next i
			Return result
		End Function

		Public Function GetBiases() As Double()
			Dim result(numClasses - 1) As Double
			For j As Integer = 0 To numClasses - 1
				result(j) = Me.biases(j)
			Next j
			Return result
		End Function

		Private Function ComputeOutputs(ByVal dataItem() As Double) As Double()
			' using curr weights[][] and biases[]
			' dataItem can be just x or x+y
			Dim result(numClasses - 1) As Double
			For j As Integer = 0 To numClasses - 1 ' compute z
				For i As Integer = 0 To numFeatures - 1
					result(j) += dataItem(i) * weights(i)(j)
				Next i
				result(j) += biases(j)
			Next j

			For j As Integer = 0 To numClasses - 1 ' 1 / 1 + e^-z
				result(j) = 1.0 / (1.0 + Math.Exp(-result(j)))
			Next j

			Dim sum As Double = 0.0 ' softmax scaling
			For j As Integer = 0 To numClasses - 1
				sum += result(j)
			Next j

			For j As Integer = 0 To numClasses - 1
				result(j) = result(j) / sum
			Next j

			Return result
		End Function ' ComputeOutputs

		Public Sub Train(ByVal trainData()() As Double, ByVal maxEpochs As Integer, ByVal learnRate As Double, ByVal decay As Double)
			' 'batch' approach (aggregate gradients using all data)
			Dim targets(numClasses - 1) As Double
			Dim msgInterval As Integer = maxEpochs \ 10
			Dim epoch As Integer = 0
			Do While epoch < maxEpochs
				epoch += 1

				If epoch Mod msgInterval = 0 AndAlso epoch <> maxEpochs Then
					Dim mse As Double = [Error](trainData)
					Console.Write("epoch = " & epoch)
					Console.Write("  error = " & mse.ToString("F4"))
					Dim acc As Double = Accuracy(trainData)
					Console.WriteLine("  accuracy = " & acc.ToString("F4"))
				End If

				Dim weightGrads()() As Double = MakeMatrix(numFeatures, numClasses)
				Dim biasGrads(numClasses - 1) As Double

				' compute all weight gradients, (all classes, all inputs)
				For j As Integer = 0 To numClasses - 1
					For i As Integer = 0 To numFeatures - 1
						For r As Integer = 0 To trainData.Length - 1
							Dim outputs() As Double = ComputeOutputs(trainData(r))
							For k As Integer = 0 To numClasses - 1
								targets(k) = trainData(r)(numFeatures + k)
							Next k
							Dim input As Double = trainData(r)(i)
							weightGrads(i)(j) += (targets(j) - outputs(j)) * input
						Next r
					Next i
				Next j

				' compute all bias gradients (all classes, all inputs)
				For j As Integer = 0 To numClasses - 1
					For i As Integer = 0 To numFeatures - 1
						For r As Integer = 0 To trainData.Length - 1
							Dim outputs() As Double = ComputeOutputs(trainData(r))
							For k As Integer = 0 To numClasses - 1
								targets(k) = trainData(r)(numFeatures + k)
							Next k
							Dim input As Double = 1 ' 1 is a dummy input
							biasGrads(j) += (targets(j) - outputs(j)) * input
						Next r
					Next i
				Next j

				' update all weights
				For i As Integer = 0 To numFeatures - 1
					For j As Integer = 0 To numClasses - 1
						weights(i)(j) += learnRate * weightGrads(i)(j)
						weights(i)(j) *= (1 - decay) ' wt decay
					Next j
				Next i

				' update all biases
				For j As Integer = 0 To numClasses - 1
					biases(j) += learnRate * biasGrads(j)
					biases(j) *= (1 - decay)
				Next j

			Loop ' while
		End Sub ' Train



		'public void TrainStochastic(double[][] trainData, int maxEpochs, double learnRate)
		'{
		'  int epoch = 0;
		'  int[] sequence = new int[trainData.Length]; // random order
		'  for (int i = 0; i < sequence.Length; ++i)
		'    sequence[i] = i;

		'  while (epoch < maxEpochs)
		'  {
		'    ++epoch;

		'    if (epoch % 10 == 0 && epoch != maxEpochs)
		'    {
		'      double mse = Error(trainData);
		'      Console.Write("epoch = " + epoch);
		'      Console.WriteLine("  error = " + mse.ToString("F4"));
		'    }

		'    Shuffle(sequence); // process data in random order

		'    // "stochastic"/"online"/"incremental" approach
		'    // picture essential to understand
		'    for (int ti = 0; ti < trainData.Length; ++ti)
		'    {
		'      //Console.WriteLine("===================================");
		'      int r = sequence[ti]; // row index
		'      double[] computeds = ComputeOutputs(trainData[r]); // computed outputs

		'      //Console.WriteLine("Data:");
		'      //LogisticMultiProgram.ShowVector(trainData[r], 2, true);

		'      for (int j = 0; j < numClasses; ++j) // update weights for each class
		'      {
		'        double target = trainData[r][numFeatures + j]; // target 0.0 or 1.0 for curr class
		'        double computed = computeds[j];                // computed [0.0 to 1.0] for curr class

		'        for (int i = 0; i < numFeatures; ++i) // each class-specific weight
		'        {
		'          double input = trainData[r][i]; // associated curr input value
		'          //Console.WriteLine("target = " + target.ToString("F2"));
		'          //Console.WriteLine("computed = " + computed.ToString("F2"));
		'          //Console.WriteLine("input = " + input.ToString("F2"));
		'          //Console.WriteLine("old wt = " + weights[i][j].ToString("F2"));
		'          weights[i][j] += learnRate * (target - computed) * input;
		'          //Console.WriteLine("new wt = " + weights[i][j].ToString("F2"));
		'          //Console.WriteLine("");
		'        } // i = each feature
		'      } // j = each class

		'      // update biases for each class
		'      for (int j = 0; j < numClasses; ++j)
		'      {
		'        int jj = numFeatures + j;
		'        double target = trainData[r][jj];
		'        double computed = computeds[j];
		'        biases[j] += learnRate * (target - computed); // input is a dummy 1.0
		'      }

		'      //Console.WriteLine("===================================");
		'      //Console.ReadLine();
		'    } // ti
		'  } // while

		'  //return this.weights;
		'} // Train

		'private void Shuffle(int[] sequence)
		'{
		'  for (int i = 0; i < sequence.Length; ++i)
		'  {
		'    int r = rnd.Next(i, sequence.Length);
		'    int tmp = sequence[r];
		'    sequence[r] = sequence[i];
		'    sequence[i] = tmp;
		'  }
		'}

		Public Function [Error](ByVal trainData()() As Double) As Double
			' mean squared error
			Dim sumSquaredError As Double = 0.0
			For i As Integer = 0 To trainData.Length - 1 ' each training item
				Dim outputs() As Double = Me.ComputeOutputs(trainData(i))
				For j As Integer = 0 To outputs.Length - 1
					Dim jj As Integer = numFeatures + j ' column in trainData to use
					sumSquaredError += ((outputs(j) - trainData(i)(jj)) * (outputs(j) - trainData(i)(jj)))
				Next j
			Next i
			Return sumSquaredError / trainData.Length
		End Function

		Public Function Accuracy(ByVal trainData()() As Double) As Double
			' using curr wts and biases
			Dim numCorrect As Integer = 0
			Dim numWrong As Integer = 0

			For i As Integer = 0 To trainData.Length - 1
				Dim deps() As Integer = ComputeDependents(trainData(i)) ' ex: [0 1 0]
				Dim targets(numClasses - 1) As Double ' ex: [0.0 1.0 0.0]
				For j As Integer = 0 To numClasses - 1
					targets(j) = trainData(i)(numFeatures + j)
				Next j

				Dim di As Integer = MaxIndex(deps)
				Dim ti As Integer = MaxIndex(targets)
				If di = ti Then
					numCorrect += 1
				Else
					numWrong += 1
				End If
			Next i

			Return (numCorrect * 1.0) / (numWrong + numCorrect)
		End Function

		Private Shared Function MaxIndex(ByVal vector() As Double) As Integer
'INSTANT VB NOTE: The local variable maxIndex was renamed since Visual Basic will not allow local variables with the same name as their enclosing function or property:
			Dim maxIndex_Renamed As Integer = 0
			Dim maxVal As Double = vector(0)
			For i As Integer = 0 To vector.Length - 1
				If vector(i) > maxVal Then
					maxVal = vector(i)
					maxIndex_Renamed = i
				End If
			Next i
			Return maxIndex_Renamed
		End Function

		Private Shared Function MaxIndex(ByVal vector() As Integer) As Integer
'INSTANT VB NOTE: The local variable maxIndex was renamed since Visual Basic will not allow local variables with the same name as their enclosing function or property:
			Dim maxIndex_Renamed As Integer = 0
			Dim maxVal As Integer = vector(0)
			For i As Integer = 0 To vector.Length - 1
				If vector(i) > maxVal Then
					maxVal = vector(i)
					maxIndex_Renamed = i
				End If
			Next i
			Return maxIndex_Renamed
		End Function

		Private Function ComputeDependents(ByVal dataItem() As Double) As Integer()
			Dim outputs() As Double = ComputeOutputs(dataItem) ' 0.0 to 1.0
			Dim maxIndex As Integer = LogisticMulti.MaxIndex(outputs)
			Dim result(numClasses - 1) As Integer ' [0 0.. 0]
			result(maxIndex) = 1
			Return result
		End Function

	End Class ' class LogisticMulti

End Namespace ' ns
